<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2020 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.3/2027974
 */

namespace Api\Controller;

use Api\AbstractApiController;
use Application\Config\Services;
use Application\Connection\ConnectionFactory;
use Application\Model\IModelDAO;
use Application\Config\ConfigManager;
use Application\Permissions\Exception\ForbiddenException;
use Application\Permissions\Permissions;
use Events\Listener\ListenerFactory;
use Projects\Model\Project as ProjectModel;
use Workflow\Manager;
use Zend\Http\Response;
use Zend\View\Model\JsonModel;
use Workflow\Model\IWorkflow;
use Workflow\Model\Workflow;
use Workflow\Filter\Workflow as Filter;
use Record\Exception\NotFoundException;
use Application\Config\ConfigException;

/**
 * Controller for querying, creating and updating workflows.
 */
class WorkflowsController extends AbstractApiController
{
    /**
     * Create a new workflow with the provided data
     * @param mixed $data
     * @return JsonModel
     * @throws ConfigException
     */
    public function create($data)
    {
        $services = $this->services;
        try {
            $services->get(Services::CONFIG_CHECK)->enforce([IWorkflow::WORKFLOW, Permissions::AUTHENTICATED]);
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            return $this->buildErrorResponse($e);
        }
        $p4Admin  = $services->get(ConnectionFactory::P4_ADMIN);
        $defaults = Manager::getDefaultsFromConfig($this->services->get(ConfigManager::CONFIG));
        $data     = $this->applyDefaults($defaults[IWorkflow::WORKFLOW_RULES], $this->normaliseData($data));
        $filter   = new Filter($p4Admin);
        $filter->setMode(Filter::MODE_ADD)->setData($this->normaliseData($data));
        $model = $this->buildChangeResponse(new Workflow($p4Admin), $filter, ListenerFactory::WORKFLOW_CREATED);
        return $model;
    }

    /**
     * Modify part of an existing workflow, replacing the content of any field names provided with the new values
     * @param string        $workflowId     The id of the workflow being patched
     * @param array         $data           Data to patch
     * @throws ConfigException
     * @return JsonModel
     */
    public function patch($workflowId, $data)
    {
        $error = null;
        $model = null;
        try {
            $services = $this->services;
            $wfDao    = $services->get(IModelDAO::WORKFLOW_DAO);
            $services->get(Services::CONFIG_CHECK)->enforce([IWorkflow::WORKFLOW, Permissions::AUTHENTICATED]);
            $p4Admin  = $services->get(ConnectionFactory::P4_ADMIN);
            $p4_user  = $services->get(ConnectionFactory::P4_USER);
            $workflow = $wfDao->fetch($workflowId, $p4Admin);
            $defaults = Manager::getDefaultsFromConfig($this->services->get(ConfigManager::CONFIG));
            $wfArray  = $this->applyDefaults($defaults[IWorkflow::WORKFLOW_RULES], $workflow->toArray());
            $data     = $this->applyDefaults($wfArray, $data);
            if ($workflow->canEdit($p4_user)) {
                $data   = array_replace($wfArray, $this->normaliseData($data));
                $filter = new Filter($p4Admin, $workflow);
                $filter->setMode(Filter::MODE_EDIT)->setData($data);
                $model = $this->buildChangeResponse($workflow, $filter);
            } else {
                // No permission - we'll treat this as not found so as to not give an clues
                throw new NotFoundException('Cannot fetch entry. Id does not exist.');
            }
        } catch (NotFoundException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            $error = $e;
        }
        if ($error) {
            $model = $this->buildErrorResponse($error);
        }
        return $model;
    }

    /**
     * Returns the array passed with only relevant workflow values set.
     * @param array $data           data to normalise
     * @return array the normalised array
     */
    private function normaliseData(array $data)
    {
        // Only allow expected inputs
        return array_intersect_key(
            $data,
            array_flip(
                [
                    IWorkflow::SHARED,
                    IWorkflow::ON_SUBMIT,
                    IWorkflow::NAME,
                    IWorkflow::DESCRIPTION,
                    IWorkflow::OWNERS,
                    IWorkflow::END_RULES,
                    IWorkflow::AUTO_APPROVE,
                    IWorkflow::COUNTED_VOTES,
                    IWorkflow::GROUP_EXCLUSIONS,
                    IWorkflow::USER_EXCLUSIONS
                ]
            )
        );
    }

    /**
     * Applies values from defaults if not set in 'to'
     * @param array     $defaults   defaults
     * @param array     $to         apply to
     * @return array
     */
    private function applyDefaults($defaults, $to) : array
    {
        if (!isset($to[IWorkflow::GROUP_EXCLUSIONS][IWorkflow::RULE]) ||
            empty($to[IWorkflow::GROUP_EXCLUSIONS][IWorkflow::RULE])) {
            $to[IWorkflow::GROUP_EXCLUSIONS][IWorkflow::RULE] =
                $defaults[IWorkflow::GROUP_EXCLUSIONS][IWorkflow::RULE];
        }
        if (!isset($to[IWorkflow::USER_EXCLUSIONS][IWorkflow::RULE]) ||
            empty($to[IWorkflow::USER_EXCLUSIONS][IWorkflow::RULE])) {
            $to[IWorkflow::USER_EXCLUSIONS][IWorkflow::RULE] =
                $defaults[IWorkflow::USER_EXCLUSIONS][IWorkflow::RULE];
        }
        if (!isset($to[IWorkflow::COUNTED_VOTES][IWorkflow::RULE]) ||
            empty($to[IWorkflow::COUNTED_VOTES][IWorkflow::RULE])) {
            $to[IWorkflow::COUNTED_VOTES][IWorkflow::RULE] =
                $defaults[IWorkflow::COUNTED_VOTES][IWorkflow::RULE];
        }
        if (!isset($to[IWorkflow::AUTO_APPROVE][IWorkflow::RULE]) ||
            empty($to[IWorkflow::AUTO_APPROVE][IWorkflow::RULE])) {
            $to[IWorkflow::AUTO_APPROVE][IWorkflow::RULE] =
                $defaults[IWorkflow::AUTO_APPROVE][IWorkflow::RULE];
        }
        if (!isset($to[IWorkflow::END_RULES][IWorkflow::UPDATE][IWorkflow::RULE]) ||
            empty($to[IWorkflow::END_RULES][IWorkflow::UPDATE][IWorkflow::RULE])) {
            $to[IWorkflow::END_RULES][IWorkflow::UPDATE][IWorkflow::RULE] =
                $defaults[IWorkflow::END_RULES][IWorkflow::UPDATE][IWorkflow::RULE];
        }
        if (!isset($to[IWorkflow::ON_SUBMIT][IWorkflow::WITH_REVIEW][IWorkflow::RULE]) ||
            empty($to[IWorkflow::ON_SUBMIT][IWorkflow::WITH_REVIEW][IWorkflow::RULE])) {
            $to[IWorkflow::ON_SUBMIT][IWorkflow::WITH_REVIEW][IWorkflow::RULE] =
                $defaults[IWorkflow::ON_SUBMIT][IWorkflow::WITH_REVIEW][IWorkflow::RULE];
        }
        if (!isset($to[IWorkflow::ON_SUBMIT][IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE]) ||
            empty($to[IWorkflow::ON_SUBMIT][IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE])) {
            $to[IWorkflow::ON_SUBMIT][IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE] =
                $defaults[IWorkflow::ON_SUBMIT][IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE];
        }
        return $to;
    }

    /**
     * Builds a response for update requests PUT, PATCH and POST
     * @param IWorkflow     $workflow   the workflow being changed
     * @param Filter        $filter     the filter to use for values
     * @param string        $task       task performed, defaults to 'workflow updated'
     * @return JsonModel
     */
    private function buildChangeResponse($workflow, Filter $filter, $task = ListenerFactory::WORKFLOW_UPDATED)
    {
        if ($filter->isValid()) {
            $wfDao = $this->services->get(IModelDAO::WORKFLOW_DAO);
            $workflow->set($filter->getValues());
            $wfDao->save($workflow);
            $model = $this->prepareSuccessModel(['workflow' => $workflow->toArray()]);
            $this->addTask($model, $task);
        } else {
            // Unless the filter specifically set forbidden set the response code to the general 400
            if ($this->getResponse()->getStatusCode() !== Response::STATUS_CODE_403) {
                $this->getResponse()->setStatusCode(Response::STATUS_CODE_400);
            }
            $jsonModel = new JsonModel(
                [
                    'isValid'  => false,
                    'messages' => $filter->getMessages(),
                    'workflow' => null,
                ]
            );
            $model     = $this->prepareErrorModel($jsonModel);
        }
        return $model;
    }

    /**
     * Delete an existing workflow
     * @param mixed $workflowId
     * @return mixed|null|JsonModel
     * @throws \Exception
     */
    public function delete($workflowId)
    {
        $error = null;
        $model = null;
        try {
            $services   = $this->services;
            $projectDAO = $this->services->get(IModelDAO::PROJECT_DAO);
            $wfDao      = $services->get(IModelDAO::WORKFLOW_DAO);
            $services->get(Services::CONFIG_CHECK)->enforce([IWorkflow::WORKFLOW, Permissions::AUTHENTICATED]);
            $p4Admin  = $services->get(ConnectionFactory::P4_ADMIN);
            $p4_user  = $services->get(ConnectionFactory::P4_USER);
            $workflow = $wfDao->fetch($workflowId, $p4Admin);
            if ($workflow->canEdit($p4_user)) {
                $projects   = $projectDAO->fetchAll([ProjectModel::FIELD_WORKFLOW => (string) $workflowId], $p4Admin);
                $allCount   = count($projects);
                $translator = $services->get('translator');
                if (!$projects || $allCount === 0) {
                    $wfDao->delete($workflow);
                    $model = new JsonModel(
                        [
                            'isValid'  => true,
                            'messages' => [$translator->t('Workflow [%s] was deleted', [$workflowId])],
                        ]
                    );
                } else {
                    $projects = $services->get('projects_filter')->filter($projects);
                    $inUseOn  = array_map(
                        function ($project) {
                            return $project->getName();
                        },
                        $projects->getArrayCopy()
                    );
                    $inUseCount = count($inUseOn);
                    if ($inUseCount === 0) {
                        $message = $translator->t('Cannot delete workflow [%s], it is in use', [$workflowId]);
                    } else {
                        $andOthers = $inUseCount === $allCount ? '' : ' and others';
                        $message   = $translator->tp(
                            'Cannot delete workflow [%s], it is in use on project [%s]' . $andOthers,
                            'Cannot delete workflow [%s], it is in use on projects [%s]' . $andOthers,
                            count($inUseOn),
                            [$workflowId, implode(', ', $inUseOn)]
                        );
                    }
                    $model = new JsonModel(
                        [
                            'isValid'  => false,
                            'messages' => [$message],
                        ]
                    );
                }
            } else {
                // No permission - we'll treat this as not found so as to not give an clues
                throw new NotFoundException('Cannot fetch entry. Id does not exist.');
            }
        } catch (NotFoundException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            $error = $e;
        }
        if ($error) {
            $model = $this->buildErrorResponse($error);
        }
        return $model;
    }

    /**
     * Build a response based on an exception
     * @param mixed $error  the exception
     * @return JsonModel
     */
    private function buildErrorResponse($error)
    {
        return new JsonModel(
            [
                'isValid'  => false,
                'messages' => [$error->getMessage()],
            ]
        );
    }

    /**
     * Replace a workflow with the provided values
     * @param string        $workflowId     The id of the workflow being updated
     * @param array         $data           Data to patch
     * @return JsonModel
     * @throws ConfigException
     */
    public function update($workflowId, $data)
    {
        $error = null;
        $model = null;
        try {
            $services = $this->services;
            $wfDao    = $services->get(IModelDAO::WORKFLOW_DAO);
            $services->get(Services::CONFIG_CHECK)->enforce([IWorkflow::WORKFLOW, Permissions::AUTHENTICATED]);
            $p4Admin  = $services->get(ConnectionFactory::P4_ADMIN);
            $p4_user  = $services->get(ConnectionFactory::P4_USER);
            $manager  = $services->get(Services::WORKFLOW_MANAGER);
            $workflow = $wfDao->fetch($workflowId, $p4Admin);
            $data     = $this->normaliseData($data);
            // PUT should have provided all values the data just replaces defaults
            $data = $this->applyDefaults($manager->getDefaults()[IWorkflow::WORKFLOW_RULES], $data);
            if ($workflow->canEdit($p4_user)) {
                $filter = new Filter($p4Admin, $workflow);
                $filter->setMode(Filter::MODE_EDIT)->setData($data);
                $model = $this->buildChangeResponse($workflow, $filter);
            } else {
                // No permission - we'll treat this as not found so as to not give an clues
                throw new NotFoundException('Cannot fetch entry. Id does not exist.');
            }
        } catch (NotFoundException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            $error = $e;
        }
        if ($error) {
            $model = $this->buildErrorResponse($error);
        }
        return $model;
    }

    /**
     * Adds a task to the queue manager for an action
     * @param JsonModel     $workflow   the workflow model in JSON form
     * @param string        $task       task performed, defaults to 'workflow updated'
     */
    private function addTask(JsonModel $workflow, $task = ListenerFactory::WORKFLOW_UPDATED)
    {
        $this->services->get('queue')->addTask(
            $task,
            $workflow->getVariable(Workflow::WORKFLOW)['id'],
            [
                'user' => $this->services->get(ConnectionFactory::P4_USER)->getUser()
            ]
        );
    }

    /**
     * Get a list of known workflows, with an option to limit the fields returned
     * @return JsonModel
     */
    public function getList()
    {
        $services = $this->services;
        $wfDao    = $services->get(IModelDAO::WORKFLOW_DAO);
        try {
            $services->get(Services::CONFIG_CHECK)->enforce(IWorkflow::WORKFLOW);
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            return $this->buildErrorResponse($e);
        }
        $fields  = $this->getRequest()->getQuery(self::FIELDS);
        $noCache = $this->getRequest()->getQuery(self::NO_CACHE);
        $p4Admin = $services->get(ConnectionFactory::P4_ADMIN);
        $options = [];
        if (isset($noCache) && $noCache === 'true') {
            $options[IModelDAO::FETCH_NO_CACHE] = true;
        }
        $workflows        = $wfDao->fetchAll($options, $p4Admin);
        $limitedWorkflows = [];

        foreach ($workflows->toArray() as $workflow) {
            $limitedWorkflows[] = $this->limitEntityFields($workflow, $fields);
        }
        return $this->prepareSuccessModel(new JsonModel(['workflows' => $limitedWorkflows]));
    }

    /**
     * Gets a specific workflow by its id
     * @param string $workflowId the id of the workflow
     * @return JsonModel
     */
    public function get($workflowId)
    {
        $workflow = null;
        $error    = null;
        try {
            $services = $this->services;
            $wfDao    = $services->get(IModelDAO::WORKFLOW_DAO);
            $services->get(Services::CONFIG_CHECK)->enforce(IWorkflow::WORKFLOW);
            $fields   = $this->getRequest()->getQuery(self::FIELDS);
            $p4Admin  = $services->get(ConnectionFactory::P4_ADMIN);
            $workflow = $wfDao->fetch($workflowId, $p4Admin);
        } catch (NotFoundException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            $error = $e;
        }
        if ($workflow) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_200);
            return $this->prepareSuccessModel(
                new JsonModel(
                    ['workflow' => $this->limitEntityFields($workflow->toArray(), $fields)]
                )
            );
        } else {
            return $this->buildErrorResponse($error);
        }
    }
}
